==> v10.5.3.zip: brewman/brewman/brewman/routers/purchase_return.py <==
import uuid

from datetime import datetime
from decimal import Decimal
from typing import Dict, List

import brewman.schemas.input as schema_in
import brewman.schemas.voucher as output

from fastapi import APIRouter, Depends, File, HTTPException, Request, Security, status
from sqlalchemy import distinct, or_, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..core.session import get_date, set_date
from ..db.session import SessionFuture
from ..models.account_base import AccountBase
from ..models.batch import Batch
from ..models.inventory import Inventory
from ..models.journal import Journal
from ..models.product import Product
from ..models.stock_keeping_unit import StockKeepingUnit
from ..models.validations import check_duplicate_batches, check_journals_are_valid
from ..models.voucher import Voucher
from ..models.voucher_type import VoucherType
from ..schemas.blank_voucher_info import BlankVoucherInfo
from ..schemas.inventory import Inventory as InventorySchema
from ..schemas.user import UserToken
from . import get_lock_info
from .db_image import save_files, update_files
from .voucher import (
    blank_voucher,
    check_voucher_edit_allowed,
    get_batch_quantity,
    voucher_info,
)


router = APIRouter()


@router.post("", response_model=output.Voucher)
def save_route(
    request: Request,
    data: schema_in.PurchaseIn = Depends(schema_in.PurchaseIn.load_form),
    i: List[bytes] = File(None),
    t: List[bytes] = File(None),
    user: UserToken = Security(get_user, scopes=["purchase-return"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = save(data, user, db)
            save_inventories(item, data.inventories, db)
            check_duplicate_batches(item)
            save_journals(item, data.vendor, db)
            check_journals_are_valid(item)
            save_files(item.id, i, t, db)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            info = voucher_info(item, db)
            return info
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def save(data: schema_in.PurchaseIn, user: UserToken, db: Session) -> Voucher:
    product_accounts = (
        select(Product.account_id)
        .join(Product.skus)
        .where(StockKeepingUnit.id.in_([i.batch.sku.id_ for i in data.inventories]))
    )
    account_types = (
        db.execute(
            select(distinct(AccountBase.type_id)).where(
                or_(
                    AccountBase.id == data.vendor.id_,
                    AccountBase.id.in_(product_accounts),
                )
            )
        )
        .scalars()
        .all()
    )
    allowed, message = get_lock_info([data.date_], VoucherType[data.type_.replace(" ", "_").upper()], account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )
    voucher = Voucher(
        date=data.date_,
        narration=data.narration,
        is_starred=data.is_starred,
        user_id=user.id_,
        voucher_type=VoucherType[data.type_.replace(" ", "_").upper()],
    )
    db.add(voucher)
    return voucher


def save_inventories(voucher: Voucher, inventories: List[InventorySchema], db: Session) -> None:
    for d_item in inventories:
        batch = db.execute(select(Batch).where(Batch.id == d_item.batch.id_)).scalar_one()

        if d_item.quantity > batch.quantity_remaining:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"{batch.sku.product.name} ({batch.sku.units}) stock is {batch.quantity_remaining}",
            )
        if batch.name > voucher.date:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"{batch.sku.product.name} ({batch.sku.units}) "
                f"was purchased on {batch.name.strftime('%d-%b-%Y')}",
            )

        batch.quantity_remaining -= d_item.quantity

        item = Inventory(
            quantity=d_item.quantity,
            rate=batch.rate,
            tax=batch.tax,
            discount=batch.discount,
            batch=batch,
        )
        voucher.inventories.append(item)
        db.add(item)


def save_journals(voucher: Voucher, ven: schema_in.AccountLink, db: Session) -> None:
    vendor = db.execute(select(AccountBase).where(AccountBase.id == ven.id_)).scalar_one()
    journals: Dict[uuid.UUID, Journal] = {}
    amount = Decimal(0)
    for v_item in voucher.inventories:
        account_id, cc_id = db.execute(
            select(AccountBase.id, AccountBase.cost_centre_id)
            .join(AccountBase.products)
            .join(Product.skus)
            .where(StockKeepingUnit.id == v_item.batch.sku.id)
        ).one()
        amount += round(v_item.amount, 2)
        if account_id in journals:
            journals[account_id].amount += round(v_item.amount, 2)
        else:
            journals[account_id] = Journal(
                debit=-1,
                cost_centre_id=cc_id,
                account_id=account_id,
                amount=round(v_item.amount, 2),
            )
    journals[vendor.id] = Journal(
        debit=1,
        cost_centre_id=vendor.cost_centre_id,
        account_id=vendor.id,
        amount=amount,
    )
    for item in journals.values():
        voucher.journals.append(item)
        db.add(item)


@router.put("/{id_}", response_model=output.Voucher)
def update_route(
    id_: uuid.UUID,
    request: Request,
    data: schema_in.PurchaseIn = Depends(schema_in.PurchaseIn.load_form),
    i: List[bytes] = File(None),
    t: List[bytes] = File(None),
    user: UserToken = Security(get_user, scopes=["purchase-return"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = update_voucher(id_, data, user, db)
            update_inventory(item, data.inventories, db)
            check_duplicate_batches(item)
            update_journals(item, data.vendor, db)
            check_journals_are_valid(item)
            update_files(item.id, data.files, i, t, db)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            return voucher_info(item, db)
    except SQLAlchemyError as e:
        db.rollback()
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )
    except Exception:
        db.rollback()
        raise


def update_voucher(id_: uuid.UUID, data: schema_in.PurchaseIn, user: UserToken, db: Session) -> Voucher:
    voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
    product_accounts = (
        select(Product.account_id)
        .join(Product.skus)
        .where(StockKeepingUnit.id.in_([i.batch.sku.id_ for i in data.inventories]))
    )
    account_types = (
        db.execute(
            select(distinct(AccountBase.type_id)).where(
                or_(
                    AccountBase.id.in_([data.vendor.id_] + [vj.account_id for vj in voucher.journals]),
                    AccountBase.id.in_(product_accounts),
                )
            )
        )
        .scalars()
        .all()
    )
    allowed, message = get_lock_info([voucher.date, data.date_], voucher.voucher_type, account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )
    check_voucher_edit_allowed(voucher, user)
    voucher.date = data.date_
    voucher.is_starred = data.is_starred
    voucher.narration = data.narration
    voucher.user_id = user.id_
    voucher.posted = False
    voucher.last_edit_date = datetime.utcnow()
    return voucher


def update_inventory(voucher: Voucher, new_inventories: List[InventorySchema], db: Session) -> None:
    for it in range(len(voucher.inventories), 0, -1):
        item = voucher.inventories[it - 1]
        batch = db.execute(select(Batch).where(Batch.id == item.batch_id)).scalar_one()
        batch_quantity = get_batch_quantity(item.batch_id, voucher.id, db)
        index = next(
            (idx for (idx, d) in enumerate(new_inventories) if d.id_ == item.id and d.batch.id_ == item.batch.id), None
        )
        if index is not None:
            new_inventory = new_inventories.pop(index)
            if new_inventory.quantity > batch_quantity:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=f"Maximum quantity available for "
                    f"{batch.sku.product.name} ({batch.sku.units}) is {batch_quantity}",
                )
            if batch.name > voucher.date:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=f"Voucher cannot be before {item.batch.name.strftime('%d-%b-%Y')}",
                )
            item.batch.quantity_remaining = batch_quantity - new_inventory.quantity
            item.quantity = new_inventory.quantity
            item.rate = batch.rate
            item.tax = batch.tax
            item.discount = batch.discount
        else:
            item.batch.quantity_remaining = batch_quantity
            voucher.inventories.remove(item)
            db.delete(item)
    save_inventories(voucher, new_inventories, db)


def update_journals(voucher: Voucher, ven: schema_in.AccountLink, db: Session) -> None:
    vendor = db.execute(select(AccountBase).where(AccountBase.id == ven.id_)).scalar_one()
    journals: Dict[uuid.UUID, Journal] = {}
    amount = Decimal(0)
    for v_item in voucher.inventories:
        account_id, cc_id = db.execute(
            select(AccountBase.id, AccountBase.cost_centre_id)
            .join(AccountBase.products)
            .join(Product.skus)
            .where(StockKeepingUnit.id == v_item.batch.sku.id)
        ).one()
        amount += v_item.amount
        if account_id in journals:
            journals[account_id].amount += v_item.amount
        else:
            journals[account_id] = Journal(
                debit=-1,
                cost_centre_id=cc_id,
                account_id=account_id,
                amount=v_item.amount,  # type: ignore[arg-type]
            )
    journals[vendor.id] = Journal(
        debit=1,
        cost_centre_id=vendor.cost_centre_id,
        account_id=vendor.id,
        amount=amount,
    )
    for i in range(len(voucher.journals), 0, -1):
        item = voucher.journals[i - 1]
        if item.account_id in journals:
            item.debit = journals[item.account_id].debit
            item.amount = round(journals[item.account_id].amount, 2)
            item.cost_centre_id = journals[item.account_id].cost_centre_id
            del journals[item.account_id]
        else:
            db.delete(item)
            voucher.journals.remove(item)
    for item in journals.values():
        item.amount = round(item.amount, 2)
        voucher.journals.append(item)
        db.add(item)


@router.get("/{id_}", response_model=output.Voucher)
def get_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["purchase-return"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
            return voucher_info(item, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.get("", response_model=output.Voucher)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["purchase-return"]),
) -> output.Voucher:
    additional_info = BlankVoucherInfo(date=get_date(request.session), type=VoucherType.PURCHASE_RETURN)
    with SessionFuture() as db:
        return blank_voucher(additional_info, db)

==> v11.1.5.zip: brewman/brewman/brewman/routers/purchase_return.py <==
import uuid

from datetime import datetime
from decimal import Decimal

import brewman.schemas.input as schema_in
import brewman.schemas.voucher as output

from fastapi import APIRouter, Depends, File, HTTPException, Request, Security, status
from sqlalchemy import distinct, or_, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..core.session import get_date, set_date
from ..db.session import SessionFuture
from ..models.account_base import AccountBase
from ..models.batch import Batch
from ..models.inventory import Inventory
from ..models.journal import Journal
from ..models.product import Product
from ..models.stock_keeping_unit import StockKeepingUnit
from ..models.validations import check_duplicate_batches, check_journals_are_valid
from ..models.voucher import Voucher
from ..models.voucher_type import VoucherType
from ..schemas.blank_voucher_info import BlankVoucherInfo
from ..schemas.inventory import Inventory as InventorySchema
from ..schemas.user import UserToken
from . import get_lock_info
from .db_image import save_files, update_files
from .voucher import (
    blank_voucher,
    check_voucher_edit_allowed,
    get_batch_quantity,
    voucher_info,
)


router = APIRouter()


@router.post("", response_model=output.Voucher)
def save_route(
    request: Request,
    data: schema_in.PurchaseIn = Depends(schema_in.PurchaseIn.load_form),
    i: list[bytes] = File(None),
    t: list[bytes] = File(None),
    user: UserToken = Security(get_user, scopes=["purchase-return"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = save(data, user, db)
            save_inventories(item, data.inventories, db)
            check_duplicate_batches(item)
            save_journals(item, data.vendor, db)
            check_journals_are_valid(item)
            save_files(item.id, i, t, db)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            info = voucher_info(item, db)
            return info
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def save(data: schema_in.PurchaseIn, user: UserToken, db: Session) -> Voucher:
    product_accounts = (
        select(Product.account_id)
        .join(Product.skus)
        .where(StockKeepingUnit.id.in_([i.batch.sku.id_ for i in data.inventories]))
    )
    account_types: list[int] = list(
        db.execute(  # ignore[assignment]
            select(distinct(AccountBase.type_id)).where(
                or_(
                    AccountBase.id == data.vendor.id_,
                    AccountBase.id.in_(product_accounts),
                )
            )
        )
        .scalars()
        .all()
    )
    allowed, message = get_lock_info([data.date_], VoucherType[data.type_.replace(" ", "_").upper()], account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )
    voucher = Voucher(
        date_=data.date_,
        narration=data.narration,
        is_starred=data.is_starred,
        user_id=user.id_,
        voucher_type=VoucherType[data.type_.replace(" ", "_").upper()],
    )
    db.add(voucher)
    return voucher


def save_inventories(voucher: Voucher, inventories: list[InventorySchema], db: Session) -> None:
    for d_item in inventories:
        batch = db.execute(select(Batch).where(Batch.id == d_item.batch.id_)).scalar_one()

        if d_item.quantity > batch.quantity_remaining:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"{batch.sku.product.name} ({batch.sku.units}) stock is {batch.quantity_remaining}",
            )
        if batch.name > voucher.date_:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"{batch.sku.product.name} ({batch.sku.units}) "
                f"was purchased on {batch.name.strftime('%d-%b-%Y')}",
            )

        batch.quantity_remaining -= d_item.quantity

        item = Inventory(
            quantity=d_item.quantity,
            rate=batch.rate,
            tax=batch.tax,
            discount=batch.discount,
            batch=batch,
        )
        voucher.inventories.append(item)
        db.add(item)


def save_journals(voucher: Voucher, ven: schema_in.AccountLink, db: Session) -> None:
    vendor = db.execute(select(AccountBase).where(AccountBase.id == ven.id_)).scalar_one()
    journals: dict[uuid.UUID, Journal] = {}
    amount = Decimal(0)
    for v_item in voucher.inventories:
        account_id, cc_id = db.execute(
            select(AccountBase.id, AccountBase.cost_centre_id)
            .join(AccountBase.products)
            .join(Product.skus)
            .where(StockKeepingUnit.id == v_item.batch.sku.id)
        ).one()
        amount += round(v_item.amount, 2)
        if account_id in journals:
            journals[account_id].amount += round(v_item.amount, 2)
        else:
            journals[account_id] = Journal(
                debit=-1,
                cost_centre_id=cc_id,
                account_id=account_id,
                amount=round(v_item.amount, 2),
            )
    journals[vendor.id] = Journal(
        debit=1,
        cost_centre_id=vendor.cost_centre_id,
        account_id=vendor.id,
        amount=amount,
    )
    for item in journals.values():
        voucher.journals.append(item)
        db.add(item)


@router.put("/{id_}", response_model=output.Voucher)
def update_route(
    id_: uuid.UUID,
    request: Request,
    data: schema_in.PurchaseIn = Depends(schema_in.PurchaseIn.load_form),
    i: list[bytes] = File(None),
    t: list[bytes] = File(None),
    user: UserToken = Security(get_user, scopes=["purchase-return"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = update_voucher(id_, data, user, db)
            update_inventory(item, data.inventories, db)
            check_duplicate_batches(item)
            update_journals(item, data.vendor, db)
            check_journals_are_valid(item)
            update_files(item.id, data.files, i, t, db)
            db.commit()
            set_date(data.date_.strftime("%d-%b-%Y"), request.session)
            return voucher_info(item, db)
    except SQLAlchemyError as e:
        db.rollback()
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )
    except Exception:
        db.rollback()
        raise


def update_voucher(id_: uuid.UUID, data: schema_in.PurchaseIn, user: UserToken, db: Session) -> Voucher:
    voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
    product_accounts = (
        select(Product.account_id)
        .join(Product.skus)
        .where(StockKeepingUnit.id.in_([i.batch.sku.id_ for i in data.inventories]))
    )
    account_types: list[int] = list(
        db.execute(
            select(distinct(AccountBase.type_id)).where(
                or_(
                    AccountBase.id.in_([data.vendor.id_] + [vj.account_id for vj in voucher.journals]),
                    AccountBase.id.in_(product_accounts),
                )
            )
        )
        .scalars()
        .all()
    )
    allowed, message = get_lock_info([voucher.date_, data.date_], voucher.voucher_type, account_types, db)
    if not allowed:
        raise HTTPException(
            status_code=status.HTTP_423_LOCKED,
            detail=message,
        )
    check_voucher_edit_allowed(voucher, user)
    voucher.date_ = data.date_
    voucher.is_starred = data.is_starred
    voucher.narration = data.narration
    voucher.user_id = user.id_
    voucher.posted = False
    voucher.last_edit_date = datetime.utcnow()
    return voucher


def update_inventory(voucher: Voucher, new_inventories: list[InventorySchema], db: Session) -> None:
    for it in range(len(voucher.inventories), 0, -1):
        item = voucher.inventories[it - 1]
        batch = db.execute(select(Batch).where(Batch.id == item.batch_id)).scalar_one()
        batch_quantity = get_batch_quantity(item.batch_id, voucher.id, db)
        index = next(
            (idx for (idx, d) in enumerate(new_inventories) if d.id_ == item.id and d.batch.id_ == item.batch.id), None
        )
        if index is not None:
            new_inventory = new_inventories.pop(index)
            if new_inventory.quantity > batch_quantity:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=f"Maximum quantity available for "
                    f"{batch.sku.product.name} ({batch.sku.units}) is {batch_quantity}",
                )
            if batch.name > voucher.date_:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=f"Voucher cannot be before {item.batch.name.strftime('%d-%b-%Y')}",
                )
            item.batch.quantity_remaining = batch_quantity - new_inventory.quantity
            item.quantity = new_inventory.quantity
            item.rate = batch.rate
            item.tax = batch.tax
            item.discount = batch.discount
        else:
            item.batch.quantity_remaining = batch_quantity
            voucher.inventories.remove(item)
            db.delete(item)
    save_inventories(voucher, new_inventories, db)


def update_journals(voucher: Voucher, ven: schema_in.AccountLink, db: Session) -> None:
    vendor = db.execute(select(AccountBase).where(AccountBase.id == ven.id_)).scalar_one()
    journals: dict[uuid.UUID, Journal] = {}
    amount = Decimal(0)
    for v_item in voucher.inventories:
        account_id, cc_id = db.execute(
            select(AccountBase.id, AccountBase.cost_centre_id)
            .join(AccountBase.products)
            .join(Product.skus)
            .where(StockKeepingUnit.id == v_item.batch.sku.id)
        ).one()
        amount += v_item.amount
        if account_id in journals:
            journals[account_id].amount += v_item.amount
        else:
            journals[account_id] = Journal(
                debit=-1,
                cost_centre_id=cc_id,
                account_id=account_id,
                amount=v_item.amount,
            )
    journals[vendor.id] = Journal(
        debit=1,
        cost_centre_id=vendor.cost_centre_id,
        account_id=vendor.id,
        amount=amount,
    )
    for i in range(len(voucher.journals), 0, -1):
        item = voucher.journals[i - 1]
        if item.account_id in journals:
            item.debit = journals[item.account_id].debit
            item.amount = round(journals[item.account_id].amount, 2)
            item.cost_centre_id = journals[item.account_id].cost_centre_id
            del journals[item.account_id]
        else:
            db.delete(item)
            voucher.journals.remove(item)
    for item in journals.values():
        item.amount = round(item.amount, 2)
        voucher.journals.append(item)
        db.add(item)


@router.get("/{id_}", response_model=output.Voucher)
def get_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["purchase-return"]),
) -> output.Voucher:
    try:
        with SessionFuture() as db:
            item: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
            return voucher_info(item, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.get("", response_model=output.Voucher)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["purchase-return"]),
) -> output.Voucher:
    additional_info = BlankVoucherInfo(date_=get_date(request.session), type_=VoucherType.PURCHASE_RETURN)  # type: ignore[arg-type]
    with SessionFuture() as db:
        return blank_voucher(additional_info, db)
